""" render the ansi

render as svg
"""
import asyncio
from os import remove
from pathlib import Path
import svgwrite
from yaml import safe_load
from PIL import Image, ImageDraw, ImageFont
from pyppeteer import launch
from ansitoimg.ansirep import AnsiBlocks, findLen

THISDIR = str(Path(__file__).resolve().parent)

# monospaced chars have a constant height and width
TEXT_HEIGHT = 15
TEXT_WIDTH = 8.7


def ansiToSVG(ansiText, fileName, theme=THISDIR + "/onedark.yml"):
	"""convert an ansi stream to svg

	Args:
		ansiText (string): ansi text to convert
		fileName (string): file path to svg to write
		theme (str, optional): file path to base24 theme to use. Defaults to "onedark.yml".
	"""
	themeData = safe_load(open(theme))
	ansiBlocks = AnsiBlocks(ansiText)
	ansiBlocks.process()
	blocks = ansiBlocks.ansiBlocks
	size = (70 * TEXT_WIDTH, TEXT_HEIGHT * ansiBlocks.height + 5)
	dwg = svgwrite.Drawing(fileName, size)
	dwg.add(dwg.rect((0, 0), size, fill="#" + themeData["base00"]))
	group = dwg.g(style=
	"font-size:14px;font-family:FiraCode NF, Fira Code, Courier New, monospace;")
	for block in blocks:
		if block.bgColour is not None:
			group.add(
			dwg.rect((block.position[0] * TEXT_WIDTH + 5,
			block.position[1] * TEXT_HEIGHT + 2.5),
			(findLen(block.text) * 9.5, TEXT_HEIGHT), fill=block.bgColour))
		style = "" if any([
		block.bold, block.italic, block.underline, block.crossedOut]) else None
		if block.bold:
			style += "font-weight: bold;"
		if block.italic:
			style += "font-style: italic;"
		if block.underline:
			style += "text-decoration: underline;"
		if block.crossedOut:
			style += "text-decoration: line-through;"
		group.add(
		dwg.text(
		block.text, insert=(block.position[0] * TEXT_WIDTH + 5,
		(block.position[1] + 1) * TEXT_HEIGHT), fill=("#" + themeData["base05"]
		if block.fgColour is None else block.fgColour), style=style))
	dwg.add(group)
	dwg.save()


def ansiToRaster(ansiText, fileName, theme=THISDIR + "/onedark.yml"):
	"""convert an ansi stream to a raster image with pillow

	Args:
		ansiText (string): ansi text to convert
		fileName (string): image file path
		theme (str, optional): file path to base24 theme to use. Defaults to "onedark.yml".
	"""
	themeData = safe_load(open(theme))
	ansiBlocks = AnsiBlocks(ansiText)
	ansiBlocks.process()
	blocks = ansiBlocks.ansiBlocks
	size = (int(70 * TEXT_WIDTH), int(TEXT_HEIGHT * ansiBlocks.height + 5))
	image = Image.new("RGB", size, "#" + themeData["base00"])
	draw = ImageDraw.Draw(image)
	# Load the fonts
	fontNormal = ImageFont.truetype(THISDIR + "/resources/FiraCode-Regular.otf",
	14)
	fontBold = ImageFont.truetype(THISDIR + "/resources/FiraCode-Bold.otf", 14)
	fontItalic = ImageFont.truetype(THISDIR + "/resources/FiraCode-Italic.otf",
	14)
	fontBoldItalic = ImageFont.truetype(
	THISDIR + "/resources/FiraCode-BoldItalic.otf", 14)
	fontEmoji = ImageFont.truetype(
	THISDIR + "/resources/TwitterColorEmoji-SVGinOT30.ttf", 14)
	# Iterate through the ansi blocks
	for block in blocks:
		posY = block.position[1] * TEXT_HEIGHT + 2.5
		if block.bgColour is not None:
			posX = block.position[0] * TEXT_WIDTH + 5
			draw.rectangle((posX, posY, posX + findLen(block.text) * 9.5,
			posY + TEXT_HEIGHT), block.bgColour)
		text = block.text
		font = fontNormal
		fill = ("#" + themeData["base05"]
		if block.fgColour is None else block.fgColour) # get the block fill colour
		if block.bold and block.italic:
			font = fontBoldItalic
		elif block.bold:
			font = fontBold
		elif block.italic:
			font = fontItalic
		index = 0
		for char in text:
			posX = (block.position[0] + index) * TEXT_WIDTH + 5
			if ord(char) > 10000: # I wish there was a better way of doing this...
				draw.text((posX, posY + 2), char, font=fontEmoji, fill=fill)
				index += 2
			else:
				draw.text((posX, posY), char if not block.crossedOut else '\u0336' + char +
				'\u0336', font=font, fill=fill)
				index += 1
			if block.underline:
				draw.line((posX, posY + TEXT_HEIGHT, posX + 9.5, posY + TEXT_HEIGHT),
				fill=fill, width=1)
	image.save(fileName)


def ansiToSVGRaster(ansiText, fileName, theme=THISDIR + "/onedark.yml"):
	"""convert an ansi stream to a raster image using pypeteer to take a
	screenshot of a generated svg (hacky but we can get coloured emoji now)

	Args:
		ansiText (string): ansi text to convert
		fileName (string): image file path
		theme (str, optional): file path to base24 theme to use. Defaults to "onedark.yml".
	"""
	ansiToSVG(ansiText, THISDIR + "/temp.svg", theme)
	ansiBlocks = AnsiBlocks(ansiText)
	ansiBlocks.process()
	size = (int(70 * TEXT_WIDTH), int(TEXT_HEIGHT * ansiBlocks.height + 5))
	asyncio.get_event_loop().run_until_complete(
	_doGrabWebpage('file:///' + THISDIR + "/temp.svg", size, fileName))
	try:
		remove(THISDIR + "/temp.svg")
	except PermissionError:
		print("Unable to clean up, manually remove temp.svg from project root " +
		"or ignore")


async def _doGrabWebpage(url, resolution, fileName):
	''' Go to a URL, with a browser with a set resolution and take a screenshot'''
	browser = await launch(
	options={'args': ['--no-sandbox', '--disable-web-security']})
	page = await browser.newPage()
	await page.setViewport({"width": resolution[0], "height": resolution[1]})
	await page.goto(url)
	await page.screenshot({'path': fileName})
	await browser.close()



""" representation and processing of an ansi stream into an 'ast' of sorts
such that it can be rendered
"""
from ansitoimg.utils import ansiColourToRGB, findLen


class AnsiBlock():
	""" represent a block of ansi text. eg \033[31mhello!\033[0m

	ansi text can have the following attributes:
	- text
	- background colour
	- foreground colour
	- absolute char offset
	- bold
	- italic
	- underline
	- crossed out
	"""
	def __init__(self, text, position, bgColour=None, fgColour=None, bold=False,
	italic=False, underline=False, crossedOut=False):
		"""Constructor

		Args:
			text (string): text content to render
			position (int, int): x, y tuple to store the absolute offset in chars
			bgColour (string, optional): background colour. Defaults to None.
			fgColour (string, optional): foreground colour. Defaults to None.
			bold (bool, optional): is text bold?. Defaults to False.
			italic (bool, optional): is text italic?. Defaults to False.
			underline (bool, optional): is text underlined?. Defaults to False.
			crossedOut (bool, optional): is text crossed out?. Defaults to False.
		"""
		self.text = text
		self.bgColour = bgColour
		self.fgColour = fgColour
		self.bold = bold
		self.italic = italic
		self.underline = underline
		self.crossedOut = crossedOut
		self.position = position


class AnsiBlocks():
	""" representation of ansi blocks

	ansi blocks has the following attributes:
	- ansi text
	- ansi blocks
	ansi blocks also holds a series of attributes that are used when processing
	- sgi buffer
	- text buffer
	- background colour
	- foreground colour
	- absolute char offset
	- bold
	- italic
	- underline
	- crossed out
	- height
	- char pointer
	"""
	def __init__(self, ansiText):
		"""Constructor

		Args:
			ansiText (string): ansi text stream to process
		"""
		# text to parse
		self.ansiText = ansiText
		# list of blocks
		self.ansiBlocks = []
		# track various attributes
		self.sgiBuffer = []
		self.textBuffer = []
		self.bold = False
		self.italic = False
		self.underline = False
		self.crossedOut = False
		self.bgColour = None
		self.fgColour = None
		self.absX = 0
		self.absY = 0

		self.height = 1
		self.pointer = 0

	def process(self):
		""" process the ansi text into a series of ansi blocks """
		while self.pointer < len(self.ansiText):
			# process sgi codes
			if self.ansiText[self.pointer] == "\033":
				self.processSgi()
			# if not an sgi code then add to the text buffer
			else:
				self.textBuffer.append(self.ansiText[self.pointer])
				self.pointer += 1
			# if there is a 'closing' sgi code, create an ansi block
			if len(self.sgiBuffer) > 0 and self.sgiBuffer[-1] in ("\033[0m", "\033[21m",
			"\033[23m", "\033[24m", "\033[29m", "\033[39m", "\033[49m"):
				self.processCloseSgi()
		# create an ansi block out of anythin left over
		if len(self.textBuffer) > 0:
			self.setAnsiBlocks("".join(self.textBuffer))

	def setAnsiBlocks(self, text):
		"""create a series of ansi blocks from the text buffer and other attributes

		Args:
			text (string): text from the buffer
		"""
		writeTxt = []
		for char in text:
			if char == "\n" or self.absX + findLen(writeTxt) > 59: # triggers a newline
				self.ansiBlocks.append(
				AnsiBlock("".join(writeTxt), (self.absX, self.absY), self.bgColour,
				self.fgColour, self.bold, self.italic))
				self.absX = 0
				self.absY += 1
				self.height += 1
				writeTxt = []
			if char != "\n":
				writeTxt.append(char)
		if findLen(writeTxt) > 0:
			self.ansiBlocks.append(
			AnsiBlock("".join(writeTxt), (self.absX, self.absY), self.bgColour,
			self.fgColour, self.bold, self.italic, self.underline, self.crossedOut))
			self.absX += findLen(writeTxt)
		self.textBuffer = []

	def processSgi(self):
		""" process an sgi code and set attributes accordingly """
		# if text is before the sgi code. create an ansi block
		if len(self.textBuffer) > 0:
			self.setAnsiBlocks("".join(self.textBuffer))
		sgiChar = []
		# 'm' is used to mark the end of an sgi code. read to the end
		while self.ansiText[self.pointer] != "m":
			sgiChar.append(self.ansiText[self.pointer])
			self.pointer += 1
		sgiChar.append(self.ansiText[self.pointer])
		self.pointer += 1
		# set attributes based on the sgi flag
		sgiFlag = "".join(sgiChar)
		if sgiFlag == "\033[1m":
			self.bold = True
		elif sgiFlag in ("\033[2m", "\033[5m", "\033[6m", "\033[7m", "\033[8m", ):
			pass
		elif sgiFlag == "\033[3m":
			self.italic = True
		elif sgiFlag == "\033[4m":
			self.underline = True
		elif sgiFlag == "\033[9m":
			self.crossedOut = True
		elif sgiFlag != "\033[39m" and sgiFlag.startswith(
		"\033[3") or sgiFlag.startswith("\033[9"):
			self.fgColour = ansiColourToRGB(sgiFlag)
		elif sgiFlag != "\033[49m" and sgiFlag.startswith(
		"\033[4") or sgiFlag.startswith("\033[10"):
			self.bgColour = ansiColourToRGB(sgiFlag)
		self.sgiBuffer.append(sgiFlag)

	def processCloseSgi(self):
		""" process a closing sgi code and create ansi blocks accordingly
		reset any attributes that need setting """
		self.setAnsiBlocks("".join(self.textBuffer))
		# reset as per sgi code
		if self.sgiBuffer[-1] in ("\033[0m", "\033[49m"):
			self.bgColour = None
		if self.sgiBuffer[-1] in ("\033[0m", "\033[39m"):
			self.fgColour = None
		if self.sgiBuffer[-1] in ("\033[0m", "\033[21m"):
			self.bold = False
		if self.sgiBuffer[-1] in ("\033[0m", "\033[23m"):
			self.italic = False
		if self.sgiBuffer[-1] in ("\033[0m", "\033[24m"):
			self.underline = False
		if self.sgiBuffer[-1] in ("\033[0m", "\033[29m"):
			self.crossedOut = False
		self.textBuffer = []
		self.sgiBuffer = []
